# Inequally yellowed nega film correction tool for ImageJ simple version
# Ver. 0.10 Interactive Dialog Version
# (c)2021 Ohnishi, Yasuo / All right reserved.

from ij import IJ, ImagePlus, CompositeImage
from ij import WindowManager as WM
from ij.io import DirectoryChooser, OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.process import ImageConverter, ImageProcessor
from loci.plugins import BF
import math
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')
import datetime
from java.awt.event import ItemListener
from java.lang import Integer

#---Convert Value in 8, 16, 32bit mode
def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = Value * 256
    elif Mode == "32-bit":
        Value = Value * 256 * 256
    else:
        Value = Value
    return Value

#------ calculating B-G mean difference routine ############

def calcMean(ip0, ip1, xStart, xEnd, nYDivide, yStart, yEnd):
#   ip0: B image / ip1: G image / xStart & xEnd: Starting & Ending Point (x) of ROI
#   nYDivide:  number of dividing zones in Y
    print ("Mean calculate routine xStart - xEnd", xStart, xEnd, "yStart - yEnd", yStart, yEnd)
    impTmp = []
    impTmp.append(ip0)
    impTmp.append(ip1)
    nYDivide = int(nYDivide) # number of dividing zones in Y
    width = ip0.getWidth() # get image width
    height = ip0.getHeight() # get image height
    pixelNum = width*height # get total pixel number
    print ("W x H (ip0)", str(width), str(height))
    if yStart == 0 and yEnd == 0:
        height2 = height
    else:
        height2 = yEnd - yStart
    zoneBreadth = height2 // nYDivide # divided Y zone height
    tmpsumDiff = [0] * nYDivide # sum of B-G differance in a zone
    n_tmpsumDiff = [0] * nYDivide # sum of pixel number in a zone
    mean2 = [0.0] * nYDivide # mean B-G differance in a zone

    for x in range(xStart, xEnd): # processing in each X row
#       ----   
        for yz in range(nYDivide): # processing in each Y zone
            for yj in range(yStart + yz * zoneBreadth, yStart + (yz + 1) * zoneBreadth - 1): 
#                if x - xStart == 1: #### for debug
#                     print("x = 1 yj =", yj)
                Px0 = impTmp[0].getPixel(x, yj) # get B pixel Value
                Px1 = impTmp[1].getPixel(x, yj) # get G pixel Value
                tmpsumDiff[yz] = tmpsumDiff[yz] + (Px0[0] - Px1[0]) # summate B-G difference in each Y Zone
                n_tmpsumDiff[yz] = n_tmpsumDiff[yz] + 1 # count pixel number in each Y Zone
#       ----   
    for yz in range(nYDivide): # calculate mean B-G difference in each Y Zone
        mean2[yz] = float(tmpsumDiff[yz]) / float(n_tmpsumDiff[yz])
#       ----   
    return mean2 # Return value: mean B-G difference in Y Zones

#-------------------------

######### Image correcting routine ################
def image_correct(impB, impB2, impG, xStart, xEnd, Mode, GmixRatio, BgainRatio, nYDivide, mean2, yStart, yEnd, cor_Opt): 
# impB: B Image impB2: B image to be corrected mean: mean differance of B-G
# xStart/xEnd: x start/end value of ROI, Mode: bit depth of image: GmixRatio: Maximum R Ch. mixing ratio
# BgainRatio: Maximum Rch. value subtracting ratio nYDivide: dividing number of ROI in Y direction
# mean2: mean B-G differance value in each divided ROIs 
    print ("Image correct routine xStart - xEnd", xStart, xEnd, "yStart - yEnd", yStart, yEnd)
    if cor_Opt == "No":
        correct_opt = "signed"
    else:
        correct_opt = "unsigned" # Only Correct if B < G  [B - G < 0]
#                  signed or unsigned
    nYDivide = int(nYDivide) # number of Y zone
    width = impB.getWidth() # get image width
    if xEnd > width:
        xEnd = width
    height = impG.getHeight() # get image height
    if yStart == 0 and yEnd == 0:
        height2 = height
    else:
        height2 = yEnd - yStart
    correctValue = []
    maxcorrectValue = 0
    mincorrectValue = 0
    maxCorV = [0.0 for yz in range(nYDivide)]
    minCorV = [0.0 for yz in range(nYDivide)]
#### Getting B-G value difference in each pixel
    tmpMxDiffBG = [[0.0 for yj in range(height2)] for xi in range(xEnd - xStart)]
    timenow = datetime.datetime.now()
    print("Start Read tmpMxDiffBG:", timenow)
    for x in range(xStart, xEnd): # process in each x col
        for y in range(yStart, yStart + height2):
#                   get blue image pixel value
            PxB = impB.getPixel(x, y) 
#                   get Green image pixel value
            PxG = impG.getPixel(x, y) 
            tmpMxDiffBG[x - xStart][y - yStart] = PxB[0] - PxG[0]   #####Correction A
                 # save B-G pixel value difference to tmpMxDiffBG
#------------------x, y roop end
    correctValueMx = [[0.0 for yz in range(nYDivide)] for xi in range(xEnd - xStart)]
    zoneBreadth = height2 // nYDivide # height of a Y zone
    print ("zoneBreadth: ", zoneBreadth)
    zoneLUT = [0] * height2 # set look up table for y -> yzone
    for y in range(yStart, yStart + height2):
        zoneLUT[y - yStart] = min (y // zoneBreadth, nYDivide - 1) # make LUT for y -> yzone ######
    timenow = datetime.datetime.now()
    print("Start Calculate correctValueMx:", timenow)
#### Calculating correcting value
    for xi in range(xEnd - xStart): # process in each x row in ROI
        for yz in range(nYDivide):
            sumtmpDiffBG = 0.0
            nr = 0.0
            if (xi == 0) and (yz == int(nYDivide /2)):
                timenow = datetime.datetime.now()
                print("Calcurate ", xi, "th correctValueMx:", timenow)
            for xi2 in range(xi-2, xi+2): # calculating back and forth 2 rows
                for yj in range(yStart + yz * zoneBreadth, yStart + (yz + 1) * zoneBreadth - 1): 
                    if (xi2 > 0) and (yj > 0) and (xi2 < (xEnd - xStart)) and (yj < yStart + height2): 
                       # summate B-G value difference in each Y zone
                        nr = nr + 1.0
                        sumtmpDiffBG = sumtmpDiffBG + tmpMxDiffBG[xi2][yj - yStart] ######
            if nr > 0.0:
                correctValueMx[xi][yz] = (sumtmpDiffBG / nr) - mean2[yz]
#       define correcting value
#           divide sum of B-G diff with pixel number and subtract normal B-G diff mean
            else: 
                correctValueMx[xi][yz] = 0.0
#            if correctValueMx[xi][yz] < 0.0: # if R mean < G mean in a Y zone then no correction
#                correctValueMx[xi][yz] = 0.0 
            if (xi % 100 == 0) and (yz == 5):
#            if (xi % 100 == 0) and (yz % 5 == 0):  ###### for debug
                 print ("xi: ", xi, "yz: ", yz, "correctValueMx: ", correctValueMx[xi][yz])
            maxcorrectValue = max(maxcorrectValue, correctValueMx[xi][yz]) # calculate max correction value
            maxCorV[yz] = max(maxCorV[yz], correctValueMx[xi][yz]) # calculate max correction value
            mincorrectValue = min(mincorrectValue, correctValueMx[xi][yz]) # calculate min correction value
            minCorV[yz] = min(minCorV[yz], correctValueMx[xi][yz]) # calculate min correction value
    print ("maxcorrectValue", maxcorrectValue)
    print ("mincorrectValue", mincorrectValue)
### value correction process
    timenow = datetime.datetime.now()
    print("Start Correction:", timenow)
    print("Start Correction height2:", height2)
    HPcc = 0.0
    MPcc = 0.0
    LPcc = 0.0
    for x in range(xStart, xEnd): # in each x row in ROI
        Hy = 0
        My = 0
        Ly = 0 
        for y in range(yStart, yStart + height2): # start R value correction
#            if x - xStart == 1: ###### for debug
#                 print("x = 1 y =", y)
            PxB= impB.getPixel(x, y)
            PxG= impG.getPixel(x, y)
            if correct_opt == "unsigned": # Only Correct if B < G  [B - G < 0]
                if correctValueMx[x - xStart][zoneLUT[y -yStart]] > 0.0: # if correction value > 0 then no correction
                    correctValueMx[x - xStart][zoneLUT[y -yStart]] = 0.0 ######
###############################
            if correctValueMx[x - xStart][zoneLUT[y -yStart]] <= 0.0:  ######
                Pcc0 = abs(correctValueMx[x - xStart][zoneLUT[y -yStart]]) / abs(mincorrectValue)  ######
            else:
                Pcc0 = abs(correctValueMx[x - xStart][zoneLUT[y -yStart]]) / abs(maxcorrectValue)  ######
#            Pcc0 = abs(correctValueMx[x - xStart][zoneLUT[y - yStart]]) / maxCorV[zoneLUT[y - yStart]] ######
#             calculate correction value ratio
            Pcc1 = Pcc0 ** (1/2.0) # gamma weighted
            Pcc = Pcc1 * GmixRatio # GmixRatio: Max G ch. mixing Ratio (like... 0.7 or 0.9)
########  define corrected pixel value in B
            Pxc = (PxG[0] * Pcc + PxB[0] * (1-Pcc) * (1+ BgainRatio * Pcc0) ) 
            if (x % 100 == 0) and (y == int(height2)/5 + yStart):
                HPcc = Pcc
                Hy = y
            elif (x % 100 == 0) and (y == int(height2)/2 + yStart):
                MPcc = Pcc
                My = y
            elif (x % 100 == 0) and (y == 4 * int(height2)/5 + yStart):
                LPcc = Pcc
                Ly = y
                print("x(",x,")  G Mixing Ratio H:", Hy, HPcc, "M: ",My, MPcc, "L: ",Ly, LPcc)
            if Pxc <= 0:
                Pxc = 0
            elif Pxc >= (256 * convValue(1, Mode) - 1):
                Pxc = (256 * convValue(1, Mode) - 1)
            impB2.getProcessor().set(x, y, int(Pxc))

#--------------------------------

def manipulation(imp0, Mode, Show, xStart, xEnd, folder, filename, GmixRatio, BgainRatio, nYDivide, smplxStart, smplxEnd, yStart, yEnd, cor_Opt):
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    n = 0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
        ip = i.getProcessor()
        n = n + 1
    impR =  impsCh[0]
    impG =  impsCh[1]
    impB =  impsCh[2]
    if Mode <> "8bit":
        IJ.saveAsTiff(impR, folder + filename +"_R.tif")
        IJ.saveAsTiff(impG, folder + filename +"_G.tif")
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
    impB2 = Duplicator().run(impB) # impB2: correcting R image
    width = impB.getWidth()
    iwidth = int(width * 0.015) # define width (15%) of sampling region for calculating mean B-G difference
    print ("xEnd - xStart:", xEnd - xStart)
    if xEnd - xStart >= 1:
        print ("smplxEnd - smplxStart:", smplxEnd - smplxStart)
        if smplxEnd - smplxStart > 0:
            print ("do mean2")
            mean2 = calcMean(impB, impG, smplxStart, smplxEnd, nYDivide, yStart, yEnd)
        else:
            mean2 = calcMean(impB, impG, xStart - iwidth, xStart - 1, nYDivide, yStart, yEnd)
        print ("do image correct")
        image_correct(impB, impB2, impG,  xStart, xEnd, Mode, GmixRatio, BgainRatio, nYDivide, mean2, yStart, yEnd, cor_Opt)
    impB2.show()
    mergeimp = RGBStackMerge.mergeChannels([impR, impG, impB2, None, None, None, None], True)
#    IJ.saveAsTiff(mergeimp, folder + filename +"_Merge_Stack.tif")
    # convert the composite image to the RGB image
    RGBStackConverter.convertToRGB(mergeimp)
#    comp = CompositeImage(mergeimp, CompositeImage.COLOR)
    IJ.saveAsTiff(mergeimp, folder + filename +"_Merge_8bit.tif")
    mergeimp.show()
#    comp.show()
    if Mode <> "8-bit":
        IJ.saveAsTiff(impB2, folder + filename +"_B2.tif")

    if Show == "Yes":
        imp0.hide()
        impR.hide()
        impB2.hide()
        impG.hide()
        impB.hide()
#        mergeimp.hide()
    imp0.close()
    impR.close()
    impB2.close()
    impG.close()
    impB.close()
#    mergeimp.close()

# ---ROI dialog-------
class RoiPreviewer(ItemListener):
    def __init__(self, imp, ckbox1, nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd):
        """
             imp: an ImagePlus
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                                 dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
        """
        self.imp = imp
        self.original_ip = imp.getProcessor().duplicate() # store a copy
        self.nfield_xStart = nfield_xStart
        self.nfield_xEnd = nfield_xEnd
        self.nfield_yStart = nfield_yStart
        self.nfield_yEnd = nfield_yEnd
        self.prv_ckBox1 = ckbox1
    

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.changeRoi()
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)
    
    def changeRoi(self):
        """ Execute the in-place scaling of the ImagePlus. """
        rx = Integer.parseInt(self.nfield_xStart.getText())
        ry = Integer.parseInt(self.nfield_yStart.getText())
        rxe = Integer.parseInt(self.nfield_xEnd.getText())
        rye = Integer.parseInt(self.nfield_yEnd.getText())
        rw = rxe - rx
        rh = rye - ry
        rx = int(rx)
        ry = int(ry)
        rw = int(rw)
        rh = int(rh)
        if rx + rw > self.imp.width:
           rw = self.imp.width - rx
        if ry + rh > self.imp.height:
           rh = self.imp.height - ry
        print (rx, ry, rw, rh)
        self.imp.setRoi(rx, ry, rw, rh)

def setRoiUI(Title, Message, yFixed, nfield_yStart, nfield_yEnd, Mode):
    imp = WM.getCurrentImage()
    impw = imp.width
    imph = imp.height
    roi = imp.getRoi()
    if roi:
        rOrgx = roi.getBounds().x
        rOrgy = roi.getBounds().y
        rOrgw = roi.getBounds().width
        rOrgh = roi.getBounds().height
    else:
        rOrgx = 0
        rOrgy = 0
        rOrgw = 0
        rOrgh = 0
    
    if not imp:
        print "Open an image first!"
        Mode = "Stop"
        return
    gd = GenericDialog(Title)
    gd.addMessage(Message)
    gd.addNumericField("Start Point (x)", rOrgx, 0)
    gd.addNumericField("End Point (x)", rOrgx + rOrgw, 0)
    if yFixed == "No":
        gd.addMessage("Y Region to be corrected")
        gd.addNumericField("Start Point (y)", rOrgy, 0)
        gd.addNumericField("End Point (y)", rOrgy + rOrgh, 0)
    else:
        gd.addMessage("Y Region to be sampled")
        gd.addMessage("Start Point (y):" + nfield_yStart.getText())
        gd.addMessage("End Point (y):" + nfield_yEnd.getText())
    gd.addCheckbox("Preview ROI", False)
    gd.addMessage("If you would like to renew preview, \n clear checkbox once and check it again.")

    # The UI elements for the above two inputs
    nfield_xStart = gd.getNumericFields().get(0) # 1st start
    nfield_xEnd = gd.getNumericFields().get(1) # 1st end
    if yFixed == "No":
        nfield_yStart = gd.getNumericFields().get(2) # 1st start
        nfield_yEnd = gd.getNumericFields().get(3) # 1st end

    checkbox1 = gd.getCheckboxes().get(0)


    previewer = RoiPreviewer(imp, checkbox1, nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd)
    checkbox1.addItemListener(previewer)

    gd.showDialog()

    if gd.wasCanceled():
        previewer.reset()
        Mode   = "Stop"
        IJ.log("User Canceled")
        checkbox1.removeItemListener(previewer)
    else:
        previewer.changeRoi()
        checkbox1.removeItemListener(previewer)
    return nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd, Mode

#--parameter dialog 
### Get parameters from the input DialogBox 
def getParameters(Mode):
    gd = GenericDialog("Inequally yellowed nega correction: (c)2021 Yasuo Ohnishi")
    gd.addMessage("Please Enter Parameters")
    gd.addNumericField("Max Mixing Ratio of G Value (0.0 to 100.0%)", 90.0, 1)
    gd.addNumericField("Max Increase Ratio of B Value (0.0 to 200.0%)", 0.0, 1)
    gd.addMessage("Vertical Dividing Zone")
    gd.addNumericField("Number of Y Dividing Zone", 25, 0)
    Show = ["Yes","No"]
    gd.addChoice("Display working windows", Show, Show[0])
    correction = ["Yes","No"]
    gd.addChoice("Correct Only B Value lesser than G", correction, correction[1])
    gd.showDialog()  # Show Dialog

#
    GmixRatio = gd.getNextNumber()
    BgainRatio = gd.getNextNumber()
    nYDivide = gd.getNextNumber()
    parm_Show   = gd.getNextChoice()
    cor_Opt   = gd.getNextChoice()
#   ---- end of parameter input 
    GmixRatio = GmixRatio / 100.0
    BgainRatio = BgainRatio / 100.0
    if GmixRatio > 1.0:
        GmixRatio = 1.0
    if BgainRatio > 1.0:
        BgainRatio = 1.0
#
    if gd.wasCanceled():  
        Mode  = "Stop"
        IJ.log("Canceled")
#
    return Mode, str(parm_Show), GmixRatio, BgainRatio, nYDivide, cor_Opt

#---main--------
meanIntensity = []
window = WM.getActiveWindow()
Mode = "Auto"
if not window:
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    filename = od.getFileName()
    if filename:
        path = folder + filename
        IJ.log(path)
        IJ.log(folder)
        IJ.log(filename)
# Always Open file with Bio-format Plugin
        imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
        imp0 = imps[0] #imp0: Original Image 
        imp0.show()
    else:
        Mode ="Stop"
else:
    imp0 = IJ.getImage()
    filename = imp0.getTitle()
    folder = IJ.getDirectory("image") # get image's original directory
    folder = IJ.getDirectory("Output_directory")
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
        Mode = "8-bit"
    elif bitdepth ==16:
        Mode = "16-bit"
    elif bitdepth ==32:
        Mode = "32-bit"
    else:
        Mode = "16-bit"
Mode = str(Mode)
if Mode <> "Stop":
    Mode, Show, GmixRatio, BgainRatio, nYDivide, cor_Opt = getParameters(Mode)
    print ("Parameters, Mode, Show, GmixRatio, BgainRatio, nYDivide", Mode, Show, GmixRatio, BgainRatio, nYDivide)
    Show = str(Show)
    print ("GmixRatio: ", GmixRatio)
    Title = "Inequally yellowed nega quick correction: Step1"
    Message = "Select correcting ROI"
    yFixed = "No"
    WM.getActiveWindow().toFront()
if Mode <> "Stop":
    nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd, Mode = setRoiUI(Title, Message, yFixed, 0, 0, Mode)
    Title = "Inequally yellowed nega quick correction: Step2"
    Message = "Select Sampling ROI"
    yFixed = "Yes"
    WM.getActiveWindow().toFront()
if Mode <> "Stop":
    nfield_smplxStart, nfield_smplxEnd, nfield_smplyStart, nfield_smplyEnd, Mode = setRoiUI(Title, Message, yFixed, nfield_yStart, nfield_yEnd, Mode)
if Mode <> "Stop":
    xStart = Integer.parseInt(nfield_xStart.getText())
    xEnd = Integer.parseInt(nfield_xEnd.getText())
    yStart = Integer.parseInt(nfield_yStart.getText())
    yEnd = Integer.parseInt(nfield_yEnd.getText())
    smplxStart = Integer.parseInt(nfield_smplxStart.getText())
    smplxEnd = Integer.parseInt(nfield_smplxEnd.getText())
    if xEnd > imp0.width:
        xEnd =  imp0.width
    if yEnd > imp0.height:
        yEnd =  imp0.height
    if smplxEnd > imp0.width:
        smplxEnd =  imp0.width
if Mode <> "Stop":
    manipulation(imp0, Mode, Show, xStart, xEnd, folder, filename, GmixRatio, BgainRatio, nYDivide, smplxStart, smplxEnd, yStart, yEnd, cor_Opt)
IJ.log("Finish")
#IJ.run("Quit")

